#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <syslog.h>
#include <stdlib.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>
#include <errno.h>

#include "had.h"

int debug_mode = 0;
int log_write_console = 0;

int fw_log_fd = -1;

/* open udp socket to 127.0.0.1:4040 to log messages to mserv */
void fw_open_log() {
    	if(fw_log_fd < 0) {
      		fw_log_fd = socket(PF_INET, SOCK_DGRAM, 0);
		if(fw_log_fd < 0) {
			log_error("could not create new socket: %s",
 					strerror(errno));
                        return;
                }

		log_debug("Created UDP logger, socket #%d", fw_log_fd);
    	}
}

/* FIXME: there "int"s _should_ be 32 bit. */
struct log_message {
	char	type[32]; /* "MESSAGE" */
	int	prio;
	int	src_id;
	char	src_info[32];
	int	length;
};

#define JHAD_GENERIC            400

static void fw_log_message(int prio, char* text) {
        if(fw_log_fd >= 0) {
		int len = sizeof(struct log_message) + strlen(text) + 1;
		struct log_message* msg = malloc(len);
      		struct sockaddr_in addr;

      		bzero(&addr, sizeof(addr));
      		addr.sin_port = htons(4040);
		addr.sin_family=AF_INET;
      		inet_pton(AF_INET, "127.0.0.1", &addr.sin_addr);

	      	bzero(msg, sizeof(msg));
		strncpy(msg->type, "MESSAGE", 32);
		msg->prio = htonl(prio);
		msg->src_id = htonl(JHAD_GENERIC);
		strncpy(msg->src_info, "HAD", 32);
		msg->length = htonl(strlen(text));
		strncpy((char *)msg + sizeof(struct log_message), text, strlen(text) + 1);

		if(sendto(fw_log_fd, (void *)msg, len, 0,
  			(struct sockaddr*)&addr, sizeof(struct sockaddr_in)) != len)
			log_warning("fw_log send failed: %s", strerror(errno));
		free(msg);
        }
}

void log_debug(const char *format, ...)
{
	va_list args;
	char buf[512];

	if (!debug_mode) return;

	va_start(args, format);
	vsnprintf(buf, sizeof(buf), format, args);
	va_end(args);

	syslog(LOG_DEBUG, "DEBUG: %s", buf);
        fw_log_message(LOG_DEBUG, buf);
	if (log_write_console) printf("DEBUG: %s\n", buf);
}

void log_info(const char *format, ...) 
{
	va_list args;
	char buf[512];

	va_start(args, format);
	vsnprintf(buf, sizeof(buf), format, args);
	va_end(args);

	syslog(LOG_INFO, "%s", buf);
        fw_log_message(LOG_INFO, buf);
	if (log_write_console) printf("INFO: %s\n", buf);
}

void log_warning(const char *format, ...) 
{
	va_list args;
	char buf[512];

	va_start(args, format);
	vsnprintf(buf, sizeof(buf), format, args);
	va_end(args);

	syslog(LOG_WARNING, "Warning: %s", buf);
        fw_log_message(LOG_WARNING, buf);
	if (log_write_console) printf("WARNING: %s\n", buf);
}

void log_error(const char *format, ...) 
{
	va_list args;
	char buf[512];

	va_start(args, format);
	vsnprintf(buf, sizeof(buf), format, args);
	va_end(args);

	syslog(LOG_ALERT, "Error: %s", buf);
        fw_log_message(LOG_ALERT, buf);
	if (log_write_console) printf("ERROR: %s\n", buf);
}

void daemonize(char* pid_file)
{
  pid_t pid;
  int nullfd;

  if ( (pid = fork()) != 0 )
    exit(0);                            /* Parent dies                  */
       
  setsid();                             /* Now we session leader        */
       
  signal(SIGQUIT, SIG_IGN);
  signal(SIGTTIN, SIG_IGN);
  signal(SIGTTOU, SIG_IGN);
  signal(SIGTSTP, SIG_IGN);

  if ( (pid = fork()) != 0 )
    exit(0);                            /* 1st child dies               */
  
  /* And second child continue */
  chdir("/");
  umask(0);

  if (!log_write_console) {
  	nullfd = open("/dev/null", O_RDWR, 0);

  	close(0);
  	close(1);
  	close(2);

  	dup2(nullfd, 0);
  	dup2(nullfd, 1);
  	dup2(nullfd, 2);

  	close(nullfd);
  }

  write_pid(pid_file);
}

#define PIDLEN 20

void write_pid( const char *pidfile )
{
  int file;
  char pid[PIDLEN];
  struct stat lstatInfo;

  snprintf(pid, PIDLEN, "%d", (int) getpid());

  if( lstat( pidfile, &lstatInfo ) == -1 )
    {
      /* If the lstat() failed for reasons other than the file not
	 existing, return a file open error */
      if( errno != ENOENT ) {
	log_error("Failed to open file %s : %s", pidfile, strerror(errno));
        return;
      }
      
      /* The file doesn't exist, create it with O_EXCL to make sure an
	 attacker can't slip in a file between the lstat() and open() */
      file = open( pidfile, O_CREAT | O_EXCL | O_RDWR, 0600 );
      if (file==-1) {
 	log_error("Failed to open file %s : %s", pidfile, strerror(errno));
        return;
     }
    }
  else
    {
      struct stat fstatInfo;

      /* Open an existing file */
      file = open( pidfile, O_RDWR );
      if (file == -1) {
	log_error("Failed open file %s : %s", pidfile, strerror(errno));
	return;
      }

      /* fstat() the opened file and check that the file mode bits and
	 inode and device match */
      if( fstat( file, &fstatInfo ) == -1 || \
	  lstatInfo.st_mode != fstatInfo.st_mode || \
	  lstatInfo.st_ino != fstatInfo.st_ino || \
	  lstatInfo.st_dev != fstatInfo.st_dev )
	{
	  close( file );
	  log_error("File %s changed, possible race attack", pidfile);
	  return;
	}

      /* If the above check was passed, we know that the lstat() and
	 fstat() were done to the same file.  Now check that there's
	 only one link, and that it's a normal file (this isn't
	 strictly necessary because the fstat() vs lstat() st_mode
	 check would also find this) */
      if( fstatInfo.st_nlink > 1 || !S_ISREG( lstatInfo.st_mode ) )
	{
	  close( file );
	  log_error("File %s must be normal file ( possible simlink attack ), exiting ...", pidfile);
          return;
	}
    }/* else */
  ftruncate( file, 0 );
  write(file, pid, strlen(pid));
  close(file);
}


